package com.google.inject.spi;

import com.google.common.base.Preconditions;
import com.google.common.collect.ImmutableList;
import com.google.inject.Module;
import java.util.List;

/**
 * Associated to a {@link Module module}, provides the module class name, the parent module {@link
 * ModuleSource source}, and the call stack that ends just before the method invocation.
 */
final class ModuleSource {

    /** The class name of module that this {@link ModuleSource} associated to. */
    private final String moduleClassName;

    /** The parent {@link ModuleSource module source}. */
    private final ModuleSource parent;

    /**
     * Permit map created by the binder that installed this module.
     *
     * <p>The permit map is a binder-scoped object, but it's saved here because these maps have to
     * outlive the binders that created them in order to be used at injector creation, and there isn't
     * a 'BinderSource' object.
     */
    private final BindingSourceRestriction.PermitMap permitMap;

    /**
     * Creates a new {@link ModuleSource} with a {@literal null} parent.
     *
     * @param moduleClass the corresponding module
     */
    ModuleSource(Class<?> moduleClass,
            BindingSourceRestriction.PermitMap permitMap) {
        this(null, moduleClass, permitMap);
    }

    /**
     * Creates a new {@link ModuleSource} Object.
     *
     * @param parent the parent module {@link ModuleSource source}
     * @param moduleClass the corresponding module
     */
    private ModuleSource(
            ModuleSource parent,
            Class<?> moduleClass,
            BindingSourceRestriction.PermitMap permitMap) {
        Preconditions.checkNotNull(moduleClass, "module cannot be null.");
        this.parent = parent;
        this.moduleClassName = moduleClass.getName();
        this.permitMap = permitMap;
    }

    /**
     * Returns the corresponding module class name.
     *
     * @see Class#getName()
     */
    String getModuleClassName() {
        return moduleClassName;
    }

    /**
     * Creates and returns a child {@link ModuleSource} corresponding to the {@link Module module}.
     *
     * @param moduleClass the corresponding module
     */
    ModuleSource createChild(Class<?> moduleClass) {
        return new ModuleSource(this, moduleClass, permitMap);
    }

    /** Returns the parent module {@link ModuleSource source}. */
    ModuleSource getParent() {
        return parent;
    }

    /**
     * Returns the class names of modules in this module source. The first element (index 0) is filled
     * by this object {@link #getModuleClassName()}. The second element is filled by the parent's
     * {@link #getModuleClassName()} and so on.
     */
    List<String> getModuleClassNames() {
        ImmutableList.Builder<String> classNames = ImmutableList.builder();
        ModuleSource current = this;
        while (current != null) {
            String className = current.moduleClassName;
            classNames.add(className);
            current = current.parent;
        }
        return classNames.build();
    }

    /**
     * Returns the size of {@link ModuleSource ModuleSources} chain (all parents) that ends at this
     * object.
     */
    int size() {
        if (parent == null) {
            return 1;
        }
        return parent.size() + 1;
    }

    /** Returns the permit map created by the binder that installed this module. */
    BindingSourceRestriction.PermitMap getPermitMap() {
        return permitMap;
    }
}
