package com.google.inject.spi;

import com.google.inject.Binding;
import com.google.inject.Provider;

/**
 * A Provider that is part of an extension which supports a custom
 * BindingTargetVisitor.
 * <p>
 * When an extension binds a provider instance, the provider can implement this
 * interface to allow users using the
 * {@link Binding#acceptTargetVisitor(BindingTargetVisitor)} method to visit a
 * custom visitor designed for that extension. A typical implementation within
 * the extension would look like
 * <pre>
 * &lt;V, B&gt; V acceptExtensionVisitor(BindingTargetVisitor&lt;B, V&gt; visitor, ProviderInstanceBinding&lt;? extends B&gt;
 * binding) {
 *   if(visitor instanceof MyCustomExtensionVisitor) {
 *     return ((MyCustomExtensionVisitor&lt;B, V&gt;)visitor).visitCustomExtension(customProperties, binding);
 *   } else {
 *     return visitor.visit(binding);
 *   }
 * }</pre>
 * 'MyCustomExtensionVisitor' in the example above would be an interface the
 * extension provides that users can implement in order to be notified of custom
 * extension information. These visitor interfaces must extend from
 * BindingTargetVisitor.
 *
 */
public interface ProviderWithExtensionVisitor<T> extends Provider<T> {

    /**
     * Instructs the extension determine if the visitor is an instance of a custom
     * extension visitor, and if so, visit it using that method. If the visitor is
     * not an instance of the custom extension visitor, this method <b>MUST</b>
     * call visitor.visit(binding).
     * <p>
     * Due to issues with generics, the type parameters of this method do not
     * relate to the type of the provider. In practice, the 'B' type will always
     * be a supertype of 'T'.
     */
    <B, V> V acceptExtensionVisitor(BindingTargetVisitor<B, V> visitor,
                                    ProviderInstanceBinding<? extends B> binding);
}
