package com.google.inject.spi;

import static com.google.common.base.Preconditions.checkNotNull;

import com.google.inject.Binder;
import com.google.inject.TypeLiteral;
import com.google.inject.internal.Errors;
import com.google.inject.matcher.Matcher;

/**
 * Registration of type converters for matching target types. Instances are created explicitly in a
 * module using {@link com.google.inject.Binder#convertToTypes(Matcher, TypeConverter)
 * convertToTypes()} statements:
 *
 * <pre>
 *     convertToTypes(Matchers.only(TypeLiteral.get(DateTime.class)), new DateTimeConverter());
 * </pre>
 *
 */
public final class TypeConverterBinding implements Element {
    private final Object source;
    private final Matcher<? super TypeLiteral<?>> typeMatcher;
    private final TypeConverter typeConverter;

    /** @since 3.0 */
    public TypeConverterBinding(
            Object source, Matcher<? super TypeLiteral<?>> typeMatcher, TypeConverter typeConverter) {
        this.source = checkNotNull(source, "source");
        this.typeMatcher = checkNotNull(typeMatcher, "typeMatcher");
        this.typeConverter = checkNotNull(typeConverter, "typeConverter");
    }

    @Override
    public Object getSource() {
        return source;
    }

    public Matcher<? super TypeLiteral<?>> getTypeMatcher() {
        return typeMatcher;
    }

    public TypeConverter getTypeConverter() {
        return typeConverter;
    }

    @Override
    public <T> T acceptVisitor(ElementVisitor<T> visitor) {
        return visitor.visit(this);
    }

    @Override
    public void applyTo(Binder binder) {
        binder.withSource(getSource()).convertToTypes(typeMatcher, typeConverter);
    }

    @Override
    public String toString() {
        return typeConverter
                + " which matches "
                + typeMatcher
                + " (bound at "
                + Errors.convert(source)
                + ")";
    }
}
