package com.google.inject;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import org.junit.Test;
import java.util.Collections;
import java.util.Objects;
import java.util.UUID;
import java.util.concurrent.Callable;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Predicate;

/**
 * Test bindings to lambdas, method references, etc.
 */
public class Java8LanguageFeatureBindingTest {

    // Some of these tests are kind of weird.
    // See https://github.com/google/guice/issues/757 for more on why they exist.

    @Test
    public void testBinding_lambdaToInterface() {
        Injector injector = Guice.createInjector(new AbstractModule() {
            @Override
            protected void configure() {
                bind(new TypeLiteral<Predicate<Object>>() {
                }).toInstance(Objects::nonNull);
            }
        });

        Predicate<Object> predicate = injector.getInstance(new Key<Predicate<Object>>() {
        });
        assertTrue(predicate.test(new Object()));
        assertFalse(predicate.test(null));
    }

    @Test
    public void testProviderMethod_returningLambda() throws Exception {
        Injector injector = Guice.createInjector(new AbstractModule() {
            @Override
            protected void configure() {
            }

            @Provides
            public Callable<String> provideCallable() {
                return () -> "foo";
            }
        });

        Callable<String> callable = injector.getInstance(new Key<Callable<String>>() {
        });
        assertEquals("foo", callable.call());
    }

    @Test
    public void testProviderMethod_containingLambda_throwingException() throws Exception {
        Injector injector = Guice.createInjector(new AbstractModule() {
            @Override
            protected void configure() {
            }

            @Provides
            public Callable<String> provideCallable() {
                if (Boolean.parseBoolean("false")) { // avoid dead code warnings
                    return () -> "foo";
                } else {
                    throw new RuntimeException("foo");
                }
            }
        });

        try {
            injector.getInstance(new Key<Callable<String>>() {
            });
        } catch (ProvisionException expected) {
            assertTrue(expected.getCause() instanceof RuntimeException);
            assertEquals("foo", expected.getCause().getMessage());
        }
    }

    @Test
    public void testProvider_usingJdk8Features() {
        try {
            Guice.createInjector(new AbstractModule() {
                @Override
                protected void configure() {
                    bind(String.class).toProvider(StringProvider.class);
                }
            });

            fail();
        } catch (CreationException expected) {
        }

        UUID uuid = UUID.randomUUID();
        Injector injector = Guice.createInjector(new AbstractModule() {
            @Override
            protected void configure() {
                bind(UUID.class).toInstance(uuid);
                bind(String.class).toProvider(StringProvider.class);
            }
        });

        assertEquals(uuid.toString(), injector.getInstance(String.class));
    }

    @Test
    public void testBinding_toProvider_lambda() {
        Injector injector = Guice.createInjector(new AbstractModule() {
            @Override
            protected void configure() {
                AtomicInteger i = new AtomicInteger();
                bind(String.class).toProvider(() -> "Hello" + i.incrementAndGet());
            }
        });

        assertEquals("Hello1", injector.getInstance(String.class));
        assertEquals("Hello2", injector.getInstance(String.class));
    }

    @Test
    public void testBinding_toProvider_methodReference() {
        Injector injector = Guice.createInjector(new AbstractModule() {
            @Override
            protected void configure() {
                bind(String.class).toProvider(Java8LanguageFeatureBindingTest.this::provideString);
            }
        });

        Provider<String> provider = injector.getProvider(String.class);
        assertEquals("Hello", provider.get());
    }

    private String provideString() {
        return "Hello";
    }

    private static final class StringProvider implements Provider<String> {
        private final UUID uuid;

        @Inject
        StringProvider(UUID uuid) {
            this.uuid = uuid;
        }

        @Override
        public String get() {
            return Collections.singleton(uuid).stream()
                    .map(UUID::toString)
                    .findFirst().get();
        }
    }
}
