package com.google.inject;

import static com.google.inject.Asserts.assertContains;
import static com.google.inject.Asserts.getDeclaringSourcePart;
import static java.lang.annotation.RetentionPolicy.RUNTIME;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.fail;

import com.google.inject.matcher.Matchers;
import com.google.inject.spi.TypeEncounter;
import com.google.inject.spi.TypeListener;

import org.junit.Before;
import org.junit.Test;
import java.lang.annotation.Retention;

public class RequestInjectionTest {

    @Before
    public void setUp() {
        HasInjections.staticField = 0;
        HasInjections.staticMethod = null;
    }

    @Test
    public void testInjectMembers() {
        final HasInjections hi = new HasInjections();

        Guice.createInjector(new AbstractModule() {
            @Override
            protected void configure() {
                bindConstant().annotatedWith(ForMethod.class).to("test");
                bindConstant().annotatedWith(ForField.class).to(5);
                requestInjection(hi);
            }
        });

        assertEquals("test", hi.instanceMethod);
        assertEquals(5, hi.instanceField);
        assertNull(HasInjections.staticMethod);
        assertEquals(0, HasInjections.staticField);
    }

    @Test
    public void testInjectStatics() throws CreationException {
        Guice.createInjector(new AbstractModule() {
            @Override
            protected void configure() {
                bindConstant().annotatedWith(ForMethod.class).to("test");
                bindConstant().annotatedWith(ForField.class).to(5);
                requestStaticInjection(HasInjections.class);
            }
        });

        assertEquals("test", HasInjections.staticMethod);
        assertEquals(5, HasInjections.staticField);
    }

    @Test
    public void testInjectMembersAndStatics() {
        final HasInjections hi = new HasInjections();

        Guice.createInjector(new AbstractModule() {
            @Override
            protected void configure() {
                bindConstant().annotatedWith(ForMethod.class).to("test");
                bindConstant().annotatedWith(ForField.class).to(5);
                requestStaticInjection(HasInjections.class);
                requestInjection(hi);
            }
        });

        assertEquals("test", hi.instanceMethod);
        assertEquals(5, hi.instanceField);
        assertEquals("test", HasInjections.staticMethod);
        assertEquals(5, HasInjections.staticField);
    }

    @Test
    public void testValidationErrorOnInjectedMembers() {
        try {
            Guice.createInjector(new AbstractModule() {
                @Override
                protected void configure() {
                    requestInjection(new NeedsRunnable());
                }
            });
        } catch (CreationException expected) {
            assertContains(expected.getMessage(),
                    "1) No implementation for java.lang.Runnable was bound",
                    "at " + NeedsRunnable.class.getName(), ".runnable(");
        }
    }

    @Test
    public void testInjectionErrorOnInjectedMembers() {
        try {
            Guice.createInjector(new AbstractModule() {
                @Override
                protected void configure() {
                    bind(Runnable.class).toProvider(() -> {
                        throw new UnsupportedOperationException();
                    });
                    requestInjection(new NeedsRunnable());
                }
            });
        } catch (CreationException expected) {
            assertContains(expected.getMessage(),
                    "1) Error in custom provider, java.lang.UnsupportedOperationException",
                    "for field at " + NeedsRunnable.class.getName() + ".runnable(",
                    "at " + getClass().getName(), getDeclaringSourcePart(getClass()));
        }
    }

    @Test
    public void testUserExceptionWhileInjectingInstance() {
        try {
            Guice.createInjector(new AbstractModule() {
                @Override
                protected void configure() {
                    requestInjection(new BlowsUpOnInject());
                }
            });
            fail();
        } catch (CreationException expected) {
            assertContains(expected.getMessage(),
                    "1) Error injecting method, java.lang.UnsupportedOperationException: Pop",
                    "at " + BlowsUpOnInject.class.getName() + ".injectInstance(");
        }
    }

    @Test
    public void testUserExceptionWhileInjectingStatically() {
        try {
            Guice.createInjector(new AbstractModule() {
                @Override
                protected void configure() {
                    requestStaticInjection(BlowsUpOnInject.class);
                }
            });
            fail();
        } catch (CreationException expected) {
            assertContains(expected.getMessage(),
                    "1) Error injecting method, java.lang.UnsupportedOperationException: Snap",
                    "at " + BlowsUpOnInject.class.getName() + ".injectStatically(");
        }
    }

    /*
     * Tests that initializables of the same instance don't clobber
     * membersInjectors in InitializableReference, so that they ultimately
     * can be requested in any order.
     */
    @Test
    public void testEarlyInjectableReferencesWithSameIdentity() {
        Injector injector = Guice.createInjector(new AbstractModule() {
            @Override
            protected void configure() {
                // Add a listener to trigger all toInstance bindings to get an Initializable.
                bindListener(Matchers.any(), new TypeListener() {
                    @Override
                    public <I> void hear(TypeLiteral<I> type, TypeEncounter<I> encounter) {
                    }
                });

                // Bind two different Keys to the IDENTITICAL object
                // ORDER MATTERS! We want the String binding to push out the Object one
                String fail = new String("better not fail!");
                bind(Object.class).toInstance(fail);
                bind(String.class).toInstance(fail);

                // Then try to inject those objects in a requestInjection,
                // letting us get into InjectableReference.get before it has
                // finished running through all its injections.
                // Each of these technically has its own InjectableReference internally.
                // ORDER MATTERS!.. because Object is injected first, that InjectableReference
                // attempts to process its members injector, but it wasn't initialized,
                // because String pushed it out of the queue!
                requestInjection(new Object() {
                    @SuppressWarnings("unused")
                    @Inject
                    Object obj;
                    @SuppressWarnings("unused")
                    @Inject
                    String str;
                });
            }
        });
    }

    @Retention(RUNTIME)
    @BindingAnnotation
    @interface ForField {
    }

    @Retention(RUNTIME)
    @BindingAnnotation
    @interface ForMethod {
    }

    static class NeedsRunnable {
        @Inject
        Runnable runnable;
    }

    static class HasInjections {

        @Inject
        @ForField
        static int staticField;
        static String staticMethod;
        @Inject
        @ForField
        int instanceField;
        String instanceMethod;

        @Inject
        static void setStaticMethod(@ForMethod String staticMethod) {
            HasInjections.staticMethod = staticMethod;
        }

        @Inject
        void setInstanceS(@ForMethod String instanceS) {
            this.instanceMethod = instanceS;
        }
    }

    static class BlowsUpOnInject {
        @Inject
        static void injectStatically() {
            throw new UnsupportedOperationException("Snap");
        }

        @Inject
        void injectInstance() {
            throw new UnsupportedOperationException("Pop");
        }
    }
}
